/*
 * MentaQueue => http://mentaqueue.soliveirajr.com
 * Copyright (C) 2012  Sergio Oliveira Jr. (sergio.oliveira.jr@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package org.mentaqueue;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.mentaqueue.util.Builder;

public class ConcurrentArrayQueue<E> implements Queue<E> {

	private final static int DEFAULT_CAPACITY = 1026 * 16;

	private final java.util.concurrent.ConcurrentLinkedQueue<E> pool;
	private final List<E> polledElements;
	private final java.util.concurrent.ArrayBlockingQueue<E> queue;
	private final Builder<E> builder;

	public ConcurrentArrayQueue(int capacity, Builder<E> builder) {
		pool = new java.util.concurrent.ConcurrentLinkedQueue<E>();
		polledElements = new LinkedList<E>();
		queue = new java.util.concurrent.ArrayBlockingQueue<E>(capacity);
		this.builder = builder;
		for (int i = 0; i < capacity; i++) {
			pool.offer(builder.newInstance());
		}
	}

	public ConcurrentArrayQueue(Builder<E> builder) {
		this(DEFAULT_CAPACITY, builder);
	}
	
	private final static <E> Builder<E> createBuilder(final Class<E> klass) {
		return new Builder<E>() {
			@Override
            public E newInstance() {
	            try {
	            	return klass.newInstance();
	            } catch(Exception e) {
	            	throw new RuntimeException(e);
	            }
            }
		};
	}
	
	public ConcurrentArrayQueue(Class<E> klass) {
		this(createBuilder(klass));
	}
	
	public ConcurrentArrayQueue(int capacity, Class<E> klass) {
		this(capacity, createBuilder(klass));
	}


	@Override
	public E nextToOffer() {
		int rem = queue.remainingCapacity();
		if (rem > 0) {
			E e = pool.poll();
			if (e == null) {
				e = builder.newInstance();
			}
			return e;
		} else {
			return null;
		}
	}

	@Override
	public void offer(E e) {
		queue.offer(e);
	}

	@Override
	public long available() {
		return queue.size();
	}

	@Override
	public E poll() {
		E returnedElem = queue.poll();
		polledElements.add(returnedElem);
		return returnedElem;
	}

	@Override
	public void done() {
		Iterator<E> iter = polledElements.iterator();
		while(iter.hasNext()) {
			pool.offer(iter.next());
			iter.remove();
		}
	}
}
